from gevent.queue import Queue, Empty
import boto3

from . import config
from .config import aws_config, IS_DEBUG
#import umysql

conn_pools = {}
boto_session = boto3.session.Session(**aws_config)

def get_dynamodb_conn():
    if(IS_DEBUG):
        return boto3.resource('dynamodb', endpoint_url='http://{endpoint}:{port}'.format(endpoint="127.0.0.1",port="8000"), aws_access_key_id=" ", aws_secret_access_key=" ", region_name="ap-south-1" )
    else:
        return boto_session.resource('dynamodb')

def get_s3_conn():
    return boto_session.resource('s3')

def get_sqs_conn():
    return boto_session.client('sqs')

def get_ses_conn():
    return boto_session.client('ses', region_name="eu-west-1")


connection_generators = {
    "dynamodb": get_dynamodb_conn,
    "s3" : get_s3_conn,
    "sqs": get_sqs_conn,
    "ses": get_ses_conn
}

connection_generators.update(**config.connection_generators)



def get_from_pool(pool_id):
    conn = None

    conn_pool = conn_pools.get(pool_id, None)
    if (conn_pool == None):
        #if no conn_pool for given pool_id create and store in conn_pools
        conn_pool = Queue()
        conn_pools[pool_id] = conn_pool

    try:
        conn = conn_pool.get(block=False)
    # print "reusing from pool"
    except Empty:
        #if queue is empty create a connection for use.
        if(pool_id in connection_generators):
            conn = connection_generators[pool_id]()

            
    return conn


def release_to_pool(conn, pool_id):
    # print "releasing to pool"
    conn_pools[pool_id].put(conn)


def use_connection_pool(**pool_args):
    def use_db_connection(func):
        def func_wrap(*args, **kwargs):

            conn_args = {}
            for pool_arg, pool_id in pool_args.items():
                conn = get_from_pool(pool_id)
                conn_args[pool_arg] = conn
            ret = None
            conn_args.update(kwargs)
            ret = func(*args, **conn_args)
            for pool_arg, pool_id in pool_args.items():
                release_to_pool(conn_args[pool_arg], pool_id)
            return ret

            # TODO: try to remove connection timeout
            # retry = 0
            # while (retry < 2):
            #     conn_args = {}
            #     for pool_arg, pool_id in pool_args.iteritems():
            #         conn = get_from_pool(pool_id)
            #         conn_args[pool_arg] = conn
            #     ret = None
            #     try:
            #         conn_args.update(kwargs)
            #         ret = func(*args, **conn_args)
            #         for pool_arg, pool_id in pool_args.iteritems():
            #             release_to_pool(conn_args[pool_arg], pool_id)
            #         return ret
            #     except Exception, e:
            #         logger.error(message=e, event="message") # dont release that stupid connections to pool, instread write destroyers for each pool
            #         logger.error(message=traceback.format_exc(), event="message")
            #
            #         retry += 1
            #         if retry >= 2:
            #             raise e
        func_wrap._original = func
        return func_wrap

    return use_db_connection



'''
Created on 19-Jan-2018

@author: abhinav
'''


from elasticsearch import Elasticsearch, RequestsHttpConnection
from requests_aws4auth.aws4auth import AWS4Auth
from .config import aws_config, es_aws_host, es_http_host
import urllib3
urllib3.disable_warnings()

es_conn = None


if(es_http_host):
    es_conn = Elasticsearch([es_http_host])

        
elif(es_aws_host):
    es_awsauth = AWS4Auth(aws_config["aws_access_key_id"] , aws_config["aws_secret_access_key"] , aws_config["region_name"], 'es')
    es_conn = Elasticsearch(
                hosts=[{'host': es_aws_host, 'port': 443}],
                http_auth=es_awsauth,
                use_ssl=True,
                verify_certs=False,
                connection_class=RequestsHttpConnection
    )

    


def create_indexes_and_mappings(index_name, index_config=None, recreate_index=False, mappings = None):
    if recreate_index:
        es_conn.indices.delete(index_name, ignore=[400, 404])
        
        
    index_config = index_config or  {
        "settings": {
            "number_of_shards": 5,
            "number_of_replicas": 1,
            "index.requests.cache.enable": False
        }
    }
    
    if(mappings):
        index_config["mappings"] = mappings
    
    return es_conn.indices.create(index=index_name, body=index_config, ignore=400)



'''
Created on 06-Jul-2018

@author: abhinav
'''

LOG_TYPE_SERVER_INFO = 1000
LOG_TYPE_PROCESSED_SQS_MESSAGE = 1001
LOG_TYPE_EXCEPTION = -1
LOG_TYPE_DEBUG = 2


TYPE_OK = 1
TYPE_ERROR = -1
TYPE_NEED_LOGIN = -2


import types
import hashlib

from pymysql import connections
from collections import OrderedDict

def jump_hash(key, num_buckets):
	b, j = -1, 0
	key = int(hashlib.md5(key.encode()).hexdigest(), 16)
	while j < num_buckets:
		b = int(j)
		key = ((key * int(2862933555777941757)) + 1) & 0xFFFFFFFFFFFFFFFF
		j = float(b + 1) * (float(1 << 31) / float((key >> 33) + 1))
	return int(b)

class MultiIterator:
	q = None

	def __init__(self):
		self.q = []

	def add(self, _iter):
		self.q.append(_iter)

	def __iter__(self):
		return self

	def __next__(self):
		while(self.q):
			try:
				return next(self.q[-1])
			except StopIteration as ex:
				self.q.pop()
		raise StopIteration

class Attribute(object):
	_type = None

	def __init__(self, _type, **kwargs):
		self._type = _type
		self.__dict__.update(kwargs)

class Model(object):
	#class level
	_attrs = None
	_pk_attrs = None

	#instance level
	__is_new = True
	_pending_updates = {}

	_inserted_id = None
	_init = False

	_json = None

	def get_custom_dict(self, path, dict_obj):

		_init = True

		class DictObj(dict):
			def __setitem__(this, k, v):
				if(not _init):
					new_path = path + "." + k
					self._pending_updates[new_path] = v
					if(isinstance(v, dict)):
						v = self.get_custom_dict(new_path, v)
				super(DictObj, this).__setitem__(k, v)

		ret = DictObj(**dict_obj)
		_init = False
		return ret

	def __setattr__(self, k, v):
		if(k in self.__class__._attrs):
			_attr_type_obj = self.__class__._attrs[k]
			if(_attr_type_obj._type == dict and isinstance(v, dict)):
				v = self.get_custom_dict(k, v)

			if(not self._init):
				self._pending_updates[k] = v
		self.__dict__[k] = v

	def to_json(self):
		self._json = _json = self._json or {}
		for k in self.__class__._attrs:
			if(k in self.__dict__):
				_json[k] = getattr(self, k , None)
		return _json

	def pk(self):
		ret = {}
		for k in self.__class__._pk_attrs.keys():
			ret[k] = getattr(self, k)
		return ret

	@classmethod
	def get_instance_from_document(cls, doc):
		ret = cls()
		ret.__is_new = False
		ret._init = True
		for k, v in cls._attrs.items():
			ret.__dict__[k] = None # set all initial attributes to None
		for k, v in doc.items():
			setattr(ret, k, v)
		ret._init = False
		return ret

	@classmethod
	def create_doc(cls, **values):
		ret = cls()
		ret._pending_updates.update(cls.get_default_values())
		for k, v in cls._attrs.items():
			ret.__dict__[k] = None # set all initial attributes to None
		for k, v in values.items(): # set the given values
			setattr(ret, k, v)
		return ret

	@classmethod
	def get_default_values(cls):
		ret = {}
		for k, v in cls._attrs.items():
			default = getattr(v, "default", None)
			if(default):
				if(isinstance(default, types.FunctionType)):
					default = default()
				ret[k] = default
		return ret

	@classmethod
	def get_shard_key_from_query(cls, _query):
		_shard_keys = cls._shard_keys_
		if(not _shard_keys):
			return ""
		ret = []
		for k in cls._shard_keys_:
			ret.append(_query[k]) # must exist
		return "__".join(ret)

	def get_shard_key(self):
		_shard_keys = self.__class__._shard_keys_
		if(not _shard_keys):
			return ""
		ret = []
		for k in _shard_keys:
			ret.append(getattr(self, k))
		return "__".join(ret)

	@classmethod
	def update(cls, _query, update_values):
		return get_shard(cls, cls.get_shard_key_from_query(_query)).update(_query, update_values) # , hint=cls._pk_attrs)

	@classmethod
	def query(cls, _query, sort=None, projection=None):
		ret = get_shard(cls, cls.get_shard_key_from_query(_query)).find(_query, projection)
		if(sort):
			ret = ret.sort(sort)
		return map(cls.get_instance_from_document, ret)

	@classmethod
	def query_multiple_shards(cls, _query, shard_keys_array, sort=None, projection=None):
		collections_map = {}
		for _shard_keys in shard_keys_array:
			_collection = None
			if(isinstance(_shard_keys, str)): # assuming single primary key
				_collection = get_shard(cls, _shard_keys)
			else:
				_collection = get_shard(cls, "__".join(_shard_keys))
			collections_map[id(_collection)] = _collection

		multi_iterator = MultiIterator()
		for _temp, _collection in collections_map.items():
			ret = _collection.find(_query, projection)
			if(sort):
				ret = ret.sort(sort)
			multi_iterator.add(ret)

		return map(cls.get_instance_from_document, multi_iterator)

	def commit(self, force=False):
		committed = False
		if(self.__is_new): # try inserting
			self.__is_new = False
			try:
				self._inserted_id = get_shard(self.__class__, self.get_shard_key()).insert_one(self._pending_updates)
				committed = True
			except DuplicateKeyError as ex:
				if(not force):
					raise(ex) # re reaise

				# try remove the primary key from pending updates
				if("_id" in self._pending_updates):
					del self._pending_updates["_id"]

				for k in list(self._pending_updates.keys()):
					if(k in self.__class__._pk_attrs):
						del self._pending_updates[k]

		if(not self.__is_new and not committed and self._pending_updates):  # try updated
			get_shard(self.__class__, self.get_shard_key()).update(self.pk(), {"$set": self._pending_updates}) # , hint=self.__class__._pk_attrs)

		self._pending_updates.clear()
		return self

	def delete(self):
		get_shard(self.__class__, self.get_shard_key()).delete_one(self.pk())


class Connection:

	collections = {}
	pool = None

	def __init__(self, host="localhost", port=3606, db_name=None, user_name=None, password=None):
		self.pool = Queue()

		self.host = host
		self.port = port
		self.db_name = db_name
		self.user_name = user_name
		self.password = password

		for cls in Model.__subclasses__():
			self.inititalize(cls)

	def get_connection(self, conn):
		ret = self.pool.get(block=False)
		cur_time = time.time()
		if(not ret):
			conn = connections.Connection(host=self.host, port=self.port, user=self.user,
											 password=self.password, database=self.db_name
										)
			last_used_timestamp = cur_time

		if(last_used_timestamp < cur_time - 10) # unused for 10 sec
			conn.ping(reconnect=True)

		return conn

	def release_connection(self, conn):
		self.pool.put( (time.time(), conn) )


	def get_shard(self, Model): # cache and get collection? needed?
		ret = self.collections.get(Model)
		if(not ret):
			self.collections[Model] = ret = self.db[Model._collection_name_]
		return ret

	'''	Add all necessary class level attributes to the model, 
		should be done before you do anything with the table '''
	def inititalize(self, Model):
		if(isinstance(Model, list)):
			for _m in Model:
				self.inititalize(_m)
			return

		Model._attrs = _model_attrs = {}
		Model._pk_attrs = None

		attrs_to_name = {}
		for k, v in Model.__dict__.items():
			if(isinstance(v, Attribute)):
				_model_attrs[k] = v
				attrs_to_name[v] = k
				attrs_to_name[k] = k # dumb , but it's one time thing and also helps converting if any given attributes as strings

		# ensure indexes created and all collections loaded to memory
		_model_indexes = getattr(Model, "_index_", [])
		Model._shard_keys_ = getattr(Model, "_shard_keys_", ())
		Model._cluster_ = getattr(Model, "_cluster_", "default")

		_pymysql_indexes_to_create = []
		for _index in _model_indexes:
			pymysql_index = []
			if(not isinstance(_index, tuple)):
				_index = (_index,)
			for _a in _index:
				_attr_name = _a
				_ordering = 1 # ascending
				if(isinstance(_a, tuple)):
					_a, _ordering = _a
				if(isinstance(_a, Attribute)):
					_attr_name = attrs_to_name[_a]

				pymysql_index.append((_attr_name, _ordering))


			if(not Model._pk_attrs): # create _pk_attrs from the first index
				Model._pk_attrs = _pk_attrs = OrderedDict()
				for i in pymysql_index: # first unique index
					_pk_attrs[i[0]] = 1

			_pymysql_indexes_to_create.append(pymysql_index)

		# set collection name to include shard_keys
		if(Model._shard_keys_):
			Model._table_name_ = Model._table_name_ + "_shard_" + ("__".join(map(lambda x: attrs_to_name[x], Model._shard_keys_)))
		# create indices in mysql

		#Left here , create table and create indexes if not exists on all shards
		for pymysql_index in _pymysql_indexes_to_create:
			self.get_shard(Model).create_index(pymysql_index, unique=True)

		if(not Model._pk_attrs): # create default _pk_attrs
			Model._pk_attrs = OrderedDict(_id=True)


_mongo_clusters = {}
def init_mongo_cluster(nodes, cluster="default"):
	if(not isinstance(nodes, list)):
		nodes = [nodes]
	db_nodes = []
	for node in nodes:
		if(isinstance(node, dict)):
			node = Connection(**node)
		db_nodes.append(node)
	_mongo_clusters[cluster] = db_nodes

def get_shard(Model, shard_key):
	cluster_nodes = _mongo_clusters[Model._cluster_]
	_node_with_data = jump_hash(shard_key, len(cluster_nodes))
	return cluster_nodes[_node_with_data].get_shard(Model)



# - Supports client-side sharding,
# - Doesn't autoshardi, you have to redistribute data if you add more db's later.

# - You can Configure mongo as a sharded cluster. Just pass the configuration server to 
# - init_mongo_cluster({"host": "", "port": 27017, "db_name": "yourdb", "user_name": "", "password": ""})



'''
Created on 22-Aug-2017

@author: abhinav
'''
import blaster
from blaster.config import IS_DEBUG

##set the configs
blaster.config.secrets = {"v0": "asdsadkjhsakjdhjksadasdasd"}


from blaster.base import start_stream_server, parse_qs_modified
import argparse 
import ujson as json
import collections
from blaster.common_funcs_and_datastructures import LRUCache, create_signed_value,\
    get_random_id, decode_signed_value,\
    set_socket_options,\
    Connection
from blaster.urllib_utils import get_data
import random
from blaster.ws_handler import WebSocketServerHandler
import time
from blaster.constants import TYPE_ERROR, TYPE_OK, LOG_TYPE_EXCEPTION
from blaster.base import route_handler, process_post_params,\
    is_server_running, server_log
import socket
from blaster.common_funcs_and_datastructures import cur_ms
import gevent
from gevent.time import sleep
import urllib
from elasticsearch.client.utils import query_params

'''
main server has all the sessions info
when other server starts , update_server_status  onto main server
create_session on main server will create on a minion server or on self
connect to the minion server for a particular session.
'''


#basic session descriptions   
#session = {"id": , "addr": , "description": ,"image": , "type": , "last_few_chats": [] }
public_sessions = LRUCache(100000)
private_sessions = LRUCache(100000)


#TODO: get the public ip address somehow
self_server_addr = None

main_server_addr = None

#relevant on main server
servers = None


TYPE_SESSION = 10
TYPE_SESSIONS = 11
TYPE_USER = 12


random_user_names = ["Pandugadu","Rgv", "Apple", "atalo ariti pandu", "erri pushpam", "naku koncham mental"]
def get_random_name(hash_val=None):
    if(hash_val):
        return random_user_names[hash(hash_val)%len(random_user_names)]
    return random.choice(random_user_names)




##app decorators

def need_user(func):
    def new_func(sock, *args , **kwargs):
        if(kwargs.get("user",None)):#atleast one has to exist
            return func(sock , *args, **kwargs)
        return  {"type": TYPE_ERROR, "error_message": "no user ? darn !."}
        
    new_func._original = getattr(func, "_original", func)
    return new_func
    
def decode_user(func):
    def new_func(sock, *args , **kwargs):
        query_params = kwargs.get("query_params",None)
        auth_key = query_params.get("auth_key", None)
        if(not auth_key):
            #try reading from headers
            headers =  kwargs.get("headers",None)
            if(headers):
                cookie = headers.get("Cookie", None)
                if(cookie):
                    cookie = cookie.decode()
                    _temp = cookie.strip('\r\n').split("; ")
                    vals = {}
                    for i in _temp:
                        vals.update(parse_qs_modified(str(i)))
                        
                    auth_key = vals.get("auth_key",None)
        
        if(auth_key):
            user_data = decode_signed_value("user_data", auth_key, max_age_days=10)
            #_type = user_data.get("type",None)
            if(user_data):
                kwargs["user"] = json.loads(user_data)
            
            else:#simple user_id
                user_id = decode_signed_value("user_key", auth_key, max_age_days=10)
                if(user_id):
                    kwargs["user"] = {"user_id":user_id, "name": get_random_name(user_id)}
            
                
            
            
        return func(sock , *args, **kwargs)
    
    new_func._original = getattr(func, "_original", func)
    return new_func


def get_internal_auth_key():
    return create_signed_value("user_data", json.dumps({"user_id": "internal", "name": "don"}))


def update_stats_on_main():
    errors = 0
    sleep(60)#initial sleep
    while(is_server_running):
        try:
            req = get_data("http://"+main_server_addr+"/update_server_stats?auth_key="+get_internal_auth_key().decode()+"&server_addr="+self_server_addr, post=json.dumps({"load": 0.0}))
            if(not req or json.loads(req.read())["type"]!=TYPE_OK):
                errors+=1
            sleep(5*60) # 5 minutes sleep
        except Exception as ex : 
            server_log(LOG_TYPE_EXCEPTION, ex="could not update on main server")
            sleep(1*60) # 1 minutes sleep


#master server 
#show all recently active sessions #lru based
@route_handler('^/get_sessions')
@decode_user
@need_user
def get_sessions(sock ,query_params=None, headers=None, post_data=None, user=None):
    recent_sessions = public_sessions.cache.items()[:-100]
    return {"type": TYPE_SESSIONS, "sessions":  [session_obj for session_id, session_obj in recent_sessions]  } 


#get individual session data
@route_handler('^/get_session/(.*)')
@decode_user
@need_user
def get_session(sock, session_id, query_params=None, headers=None, post_data=None, user=None):
    session_id = query_params.get("session_id", session_id)
    session = None
    if(session_id and session_id.startswith("private__")):
        session = private_sessions.cache.get(session_id,None)#returns empty string
    else:
        session = public_sessions.cache.get(session_id, "")#returns empty string
    
    if(session):
        return {"type": TYPE_SESSION , "session" : session }
    else:
        return create_new_session._original(sock, query_params, headers, post_data, user, session_id=session_id)
        
    return {"type": TYPE_ERROR, "error_message": "not found"} #empty body

#probably called internally between servers to main server
@route_handler('^/update_session_data')
@decode_user
@need_user
def update_session_data(sock, session_id, query_params=None, headers=None, post_data=None, user=None):
    if(user["user_id"] != "internal"):
        return {"type": TYPE_ERROR, "error_message": "not authorized"} #empty body
    if(session_id):
        public_sessions.set("session_id", json.loads(post_data)) # will update this blindly
        return {"type": TYPE_OK, "message": "session updated"} #empty body
    
    return {"type": TYPE_ERROR, "error_message": "not found"} #empty body

@route_handler('^/send_session_message/(.*)')
@decode_user
@need_user
def send_session_message(sock, session_id, query_params=None, headers=None, post_data=None, user=None):
    if(user["user_id"] != "internal"):
        return {"type": TYPE_ERROR, "error_message": "not authorized"} #empty body
    
    session_obj = session_data.exists(session_id)
    message_data = json.loads(post_data)
    if(session_obj):     
        session_connections = session_obj["connections"]
        session_connections_temp = session_obj["viewer_connections"]
        
        session_messages = session_obj["messages"]   
        session_messages.append( message_data )
                    #broadcast to all other connections in this session
        message_data_str = json.dumps(message_data)
        for conn,temp in session_connections.cache.items():
            try:
                conn.send(message_data_str)
            except Exception as ex:
                pass
        for conn,temp in session_connections_temp.cache.items():
            try:
                conn.send(message_data_str)
            except Exception as ex:
                pass

    
        return {"type": TYPE_OK, "message": "sent"} #empty body
    
    else:
        #main server            
        if(session_id and session_id.startswith("private__")):
            session = private_sessions.cache.get(session_id,None)#returns empty string
        else:
            session = public_sessions.cache.get(session_id, "")#returns empty string
        
        #call on the server and return the response
        return get_data("http://"+session["addr"]+"/send_message/"+session_id+"?auth_key="+get_internal_auth_key().decode(), post=post_data).read()
        



#update load etc
@route_handler('^/update_server_stats')
@decode_user
@need_user
def update_server_stats(sock, query_params=None, headers=None, post_data=None, user=None):
    server_addr = urllib.request.unquote(query_params.get("server_addr", None))
    if(user["user_id"] == "internal"):
        server_info = json.loads(post_data.decode())
        server_info["last_updated"] = cur_ms()
        servers.set(server_addr, server_info)
        return {"type": TYPE_OK, "message": "updated"} #empty body
    
    return {"type": TYPE_ERROR, "error_message": "not authorized"} #empty body


'''on the main server'''    
@route_handler('^/create_new_session')
@process_post_params
@decode_user
@need_user
def create_new_session(sock, query_params=None, headers=None, post_data=None, user=None, session_id=None):
    #create on some random server a
    description = query_params.get("description", None)
    image = query_params.get("image", None)
    is_public = query_params.get("is_public", "true")
    available_servers = []
    for i, server in servers.cache.items():
        if(server["load"]< 1 and server["last_updated"] > cur_ms() - 10*60*1000):#updates less than 10 minutes
            available_servers.append(i)
    
    addr = random.choice(available_servers)
    
    if(not addr):
        return {"type": TYPE_ERROR, "error_message": "no address"} #empty body
    
    if(session_id):
        pass
    elif(is_public=="true"):
        session_id = get_random_id(10)
    else:
        session_id = "private__"+get_random_id(10)
        
  
    session_obj = {"session_id": session_id, "description" : description, "image": image ,"type": 0, "addr": addr, "user": user}
    
    #update on that server
    req = get_data("http://"+addr+"/after_new_session_created?auth_key="+get_internal_auth_key().decode(), post=json.dumps(session_obj))
    if(not req or json.loads(req.read())["type"]!=TYPE_OK):
        return {"type": TYPE_ERROR, "error_message": "not found"} #empty body
        
    
    if(is_public=="true"):
        public_sessions.set(session_id ,  session_obj )
    else:
        private_sessions.set(session_id ,  session_obj )
        
    return {"type": TYPE_SESSION , "session" : session_obj }
            

#all sessions that exists on this server
session_data = LRUCache(10000)

@route_handler('^/after_new_session_created')
@decode_user
@need_user
def after_new_session_created(sock,query_params=None, headers=None, post_data=None, user=None):
    if(user["user_id"]!="internal"):
        return {"type": TYPE_ERROR, "error_message": "invalid user"} #empty body
    
    session_obj = json.loads(post_data.decode())
    session_id = session_obj["session_id"] 
    existing_session_obj = session_data.exists(session_id)
    if(not existing_session_obj):
        session_data.set(session_id , session_obj)
        #create session fields
        session_obj["connections"] = LRUCache(500)
        session_obj["viewer_connections"] = LRUCache(500)
        session_obj["messages"] = collections.deque(maxlen=100)
    return {"type": TYPE_OK, "message": "updated"} #empty body
    

@route_handler('^/create_or_update_user')
@decode_user
def create_or_update_user(sock,query_params=None, headers=None, post_data=None, user=None):
    user_name = query_params.get("name", None) or get_random_name()
    user_id = None
    if(user):
        user_id = user["user_id"]
    return {"type": TYPE_USER, "user_name": user_name ,"auth_key": create_signed_value("user_data", json.dumps({"name": user_name, "user_id" : user_id or get_random_id(10)}))}

#web socket based connection
@route_handler('^/join_session/(.*)')
@decode_user
@need_user
def join_session(sock, session_id, query_params=None, headers=None, post_data=None, user=None):
    session_id = session_id or query_params.get("session_id", None)
    session_obj = session_data.exists(session_id)
    if(not session_obj):
        return {"type": TYPE_ERROR, "error_message": "not found"} #empty body

    user_id = user["user_id"]
    session_connections = session_obj["connections"]
    session_connections_temp = session_obj["viewer_connections"]
    
    session_messages = session_obj["messages"]

    ws = WebSocketServerHandler(sock, headers)#sends handshake automatically 
    set_socket_options(sock)
    
    def delete_connection(ex):
        session_connections.delete(ws)
    
    def on_connected():
        ws.conn_obj = conn = Connection(ws, user_id)
        conn.is_viewer_list = True
        popped_connections = session_connections_temp.set(conn, True)
        for c , b in popped_connections:
            c.ws.close()
        
        #send last 100 messages
        initial_messages = json.dumps(session_messages)
        msg_to_send = {"type": 101, "payload": initial_messages}
        conn.send(json.dumps(msg_to_send), add_to_assumend_sent=False)
        
        
        
    def on_timeout():
        if(ws.conn_obj):
            ws.conn_obj.send(json.dumps({"type":0, "timestamp": int(time.time()*1000)}))
            return True
        return False
    
    def on_message():
        try:
            message_data = json.loads(ws.data)
            if(message_data.get("type", None)==1):
                if(ws.conn_obj.is_viewer_list):
                    session_connections_temp.delete(ws.conn_obj)
                    ws.conn_obj.is_viewer_list = False
                    popped_connections = session_connections.set(ws.conn_obj, True)
                    for c , b in popped_connections:
                            c.ws.close()
                
                
                message_data["src_id"] = user_id #user_id
                message_data["timestamp"] = int(time.time()*1000) #user_id
                
                session_messages.append( message_data )
                #broadcast to all other connections in this session
                message_data_str = json.dumps(message_data)
                for conn,temp in session_connections.cache.items():
                    if(conn!=ws.conn_obj):
                        try:
                            conn.send(message_data_str)
                        except Exception as ex:
                            pass
                for conn,temp in session_connections_temp.cache.items():
                    if(conn!=ws.conn_obj):
                        try:
                            conn.send(message_data_str)
                        except Exception as ex:
                            pass
        except Exception as ex:
            server_log(LOG_TYPE_EXCEPTION, exception=str(ex))
    
    ws.handleClose = delete_connection
    ws.handleConnected = on_connected
    ws.handleMessage = on_message
    ws.handleTimeout = on_timeout
     
    ws.do_handshake(headers)
    ws.start_handling()
    return





if __name__ == "__main__":
    
    parser = argparse.ArgumentParser(description='process arguments')
    parser.add_argument('--port',
                       help='host port specifically')
    
    parser.add_argument('--addr',
                       help='host name')
    
    parser.add_argument('--main_addr',
                       help='main server host name')
    
    
    
    args = parser.parse_args()
    
    port = 8081
    if(args.port):
        port = int(args.port)
    
    #TODO: get the public ip address somehow
    self_server_addr = (args.addr or socket.gethostbyname(socket.gethostname())) +":"+str(port)
    
    main_server_addr = args.main_addr or self_server_addr
    
    #relevant on main server
    servers = LRUCache(100, [(self_server_addr, {"load":0.0, "last_updated": cur_ms()})]) # {server_id : address , load: }
    

    
    gevent.spawn(update_stats_on_main)
    
    start_stream_server(port, handlers=[])
    



# from . import errors
# from . import expression
# from . import fields
# from . import helpers
# from . import indexes
# from . import json_import
# from . import model
# from . import query
# from . import table



'''
Created on 27-Feb-2017

'''
import types
from datetime import datetime

import traceback
import ujson as json
import gevent
from ..config import IS_DEBUG, sqs_url
from ..constants import LOG_TYPE_PROCESSED_SQS_MESSAGE, LOG_TYPE_SERVER_INFO, LOG_TYPE_EXCEPTION
from .. import base
from ..base import push_tasks, server_log
from ..common_funcs_and_datastructures import get_random_id,\
    cur_ms
from ..connection_pool import get_from_pool,\
    use_connection_pool



sqs_reader_greenlets = []

@use_connection_pool(queue="sqs")
def start_boto_sqs_readers(queue=None):
    if(IS_DEBUG):
        server_log(LOG_TYPE_SERVER_INFO, data="Not starting sqs readers in DEBUG mode")
        return
        
    queue_url = sqs_url
    
    def process_from_sqs():
        while(base.is_server_running):
            try:
                response = queue.receive_message(
                        QueueUrl=queue_url,
                        MessageAttributeNames = ['All'],                 
                        MaxNumberOfMessages=10,
                        VisibilityTimeout=60,
                        WaitTimeSeconds=20#long polling gevent safe
#                         ,ReceiveRequestAttemptId=''   , make it unique for each instance , probably when bootup with an ip ?
                    )
                    
                for sqs_message in response.get("Messages",[]):
                    message_payload = json.loads(sqs_message.get("Body", "{}"))
                    kwargs = message_payload.get("kwargs", {})
                    args = message_payload.get("args", [])
                    func_name = message_payload.get("func", "")
                    task_id = message_payload.get("task_id", "")
                    #TODO use task_id for logging
                    func = push_tasks.get(func_name, None)
                    if func:
                        func(*args, **kwargs)
                    else:
                        server_log(LOG_TYPE_EXCEPTION, data="Not a push task", func=str(func))

                    _temp = queue.delete_message(
                        QueueUrl=queue_url,
                        ReceiptHandle=sqs_message.get("ReceiptHandle",None)
                    )
                    
                    server_log(LOG_TYPE_PROCESSED_SQS_MESSAGE , data=json.dumps(_temp))

            except Exception as ex:
                server_log(LOG_TYPE_EXCEPTION, data=str(ex), _type="sqs_exception")
                traceback.print_exc()  
        
    for i in range(5):
        sqs_reader_greenlets.append(gevent.spawn(process_from_sqs))

    base.exit_listeners.append(wait_for_push_tasks_processing)


def push_task(func):
    if(isinstance(func, str)):
        #return decorator
        def decorator(func2):
            push_tasks[func] = func2
            return func2

        return decorator
    else:
        _original = getattr(func, "_original", func)
        push_tasks[_original.__name__] = func
        return func

@use_connection_pool(queue="sqs")
def post_a_task(func, *args, **kwargs):
    queue = kwargs.pop("queue")
    args = list(args)

    if isinstance(func, str) or isinstance(func, bytes):
        func_name = func
    elif isinstance(func, types.FunctionType):
        if hasattr(func, "_original"):
            func_name = func._original.__name__
        else:
            func_name = func.__name__
    else:
        return None

    if(IS_DEBUG):
        server_log(LOG_TYPE_SERVER_INFO, data="calling directly in DEBUG mode")
        func = push_tasks.get(func_name, None)
        func(*args, **kwargs)
        return None


    now = datetime.utcnow().isoformat()
    task_id = get_random_id()
    message_body_json = {"args": args, "kwargs": kwargs, "func": func_name, "task_id": task_id, "created_at": now}
    message_body = json.dumps(message_body_json)
    queue_url = sqs_url
    response = queue.send_message(
            QueueUrl=queue_url,
            MessageBody=message_body,
            DelaySeconds=1
    )
    return response
    


def wait_for_push_tasks_processing():
    server_log(LOG_TYPE_SERVER_INFO, data="wait called")
    gevent.joinall(sqs_reader_greenlets)
    del sqs_reader_greenlets[:]



'''
Created on 22-Aug-2017

@author: abhinav
'''

# argparse==1.2.1
# gevent==1.2.1
# greenlet==0.4.12
# protobuf==3.2.0
# boto3==1.4.4


import ujson as json
import re
import urllib.parse
import time
import gevent
import signal
import functools
import six
from gevent.server import StreamServer
from gevent.pywsgi import WSGIServer


from ..constants import LOG_TYPE_SERVER_INFO
from ..common_funcs_and_datastructures import cur_ms

is_server_running = True

request_handlers = []
exit_listeners = []


push_tasks = {} # name: func

default_wsgi_headers = [
                            ('Content-Type', 'text/html; charset=utf-8')
                    ]
default_stream_headers = [
                                'Content-Type: text/html; charset=utf-8',
                                'Access-Control-Allow-Origin: *'
                        ]



def server_log(log_type, cur_millis=None, **kwargs):
    if(not cur_millis):
        cur_millis = cur_ms()
    print(log_type , cur_millis, json.dumps(kwargs))







def write_data(socket, *_data):
    for data in _data:
        if(isinstance(data, six.text_type)):
            data = data.encode()
        n = 0
        l = len(data)
        while(n < l):
            sent = socket.send(data[n:])
            n += sent
            if(sent < 0):
                return False
    return True



class SocketDataReader():
    
    is_eof = False
    sock = None
    store = None
    _n = 0

    def __init__(self, sock):
        self.sock = sock
        self.store = bytearray()
        self._n = 0
    
    def read(self, n):
        if(self.is_eof):
            return None
        while(self._n + n > len(self.store)):
            data = self.sock.recv(1024)
            if(not data):
                self.is_eof = True
                break
            self.store.extend(data)
        ret = self.store[self._n:self._n + n]
        self._n += n
        return ret
        
            
            
    def read_line(self):
        if(self.is_eof): return None
        n = self._n
        line_found = False
        while(not line_found):
            if(n>=len(self.store)):
                data = self.sock.recv(1024)
                if(not data):
                    self.is_eof = True
                    break
                self.store.extend(data) #fetch more data
                
            if(self.store[n]==ord(b'\n')):
                line_found = True
            n+=1
        
        
        ret = self.store[self._n: n]
        self._n = n
        return ret

    
####parse query string
def parse_qs_modified(qs, keep_blank_values=0, strict_parsing=0):
    _dict = {}
    for name, value in urllib.parse.parse_qsl(qs, keep_blank_values, strict_parsing):
        if name in _dict:
            _values = _dict[name]
            if(not isinstance(_values , list)):#make it a list of values
                _dict[name] = _values = [_values]
            _values.append(value)
        else:
            _dict[name] = value
    return _dict

#processse a=b&c=d format of post data
def process_post_params(func):
    def new_func(*args, **kwargs):
        post_data = kwargs.get("post_data",None)
        query_params = kwargs["query_params"]
        if(post_data):
            query_params.update(parse_qs_modified(post_data.decode('utf-8')))
        ret = func(*args, **kwargs)
        return ret
    
    new_func._original = getattr(func, "_original", func)
    return new_func

#process data as json
def process_post_data(func):
    def new_func(*args, **kwargs):
        post_data = kwargs.get("post_data",None)
        query_params = kwargs["query_params"]
        if(post_data):
            post_data_str = post_data.decode('utf-8')
            if(post_data_str[0]=='{'):
                query_params.update(json.loads(post_data_str))
            else:
                query_params.update(parse_qs_modified(post_data_str))
        ret = func(*args, **kwargs)
        return ret
    
    new_func._original = getattr(func, "_original", func)
    return new_func



LOG_TYPE_REQUEST = 0
LOG_TYPE_REQUEST_ERROR = -1

def handle_connection(socket, address): 
    socket_data_reader = SocketDataReader(socket)
    is_okay = True#reuse the connection to handle another
    post_data = None
    while(is_okay):
        request_line = socket_data_reader.read_line()
        if(not request_line): return
        request_params = {}
        cur_millis = int(1000*time.time())
        try:
            request_type , request_path , http_version = request_line.decode("utf-8").split(" ")
            query_start_index = request_path.find("?")
            if(query_start_index!=-1):
                request_params = parse_qs_modified(request_path[query_start_index+1:])
                request_path = request_path[:query_start_index]
        
            server_log(LOG_TYPE_REQUEST , request_type=request_type , path=request_path) 
            headers = {}
            while(True):#keep reading headers
                l = socket_data_reader.read_line()
                if(l == b'\r\n'):
                    break
                if( not l):
                    return
                header_type , data = l.split(b": ", 1)
                headers[header_type.decode()] = data
            
            if(request_type == "POST" and headers.get("Content-Length", None)):
                n = int(headers.get("Content-Length", b"0").strip(b" \r\n"))
                if(n > 0):
                    post_data = bytearray()
                    while(len(post_data) < n):
                        bts = socket_data_reader.read(n)
                        if(not bts):
                            break
                        post_data.extend(bts)
                    #request_params.update(urlparse.parse_qs(str(post_data)))
            ##app specific headers
            ret = status = response_headers = body = None 
            for handler in request_handlers:
                args = handler[0].match(request_path)
                func = handler[1]            
                
                if(args != None):
                    kwargs = {}
                    kwargs["query_params"] = request_params
                    kwargs["headers"] = headers
                    if(post_data):
                        kwargs["post_data"] = post_data

                    fargs = args.groups()
                    if(fargs):
                        ret = func(socket, *fargs , **kwargs)
                        break
                    else:
                        ret = func(socket, **kwargs)
                        break
            
            #handle various return values from handler functions
            if(ret != None):
                if(isinstance(ret, tuple)):
                    l = len(ret)
                    if(l > 0):
                        body = ret[-1]
                    if(l > 1):
                        response_headers = ret[-2]
                    if(l > 2):
                        status = ret[-3]
                else:
                    if(isinstance(ret, dict) or isinstance(ret, list)):
                        #if it's object , return json encoded
                        body = json.dumps(ret)
                    else:
                        body = ret
    
            
                status = status or b"200 OK"
                response_headers = response_headers or default_stream_headers
                is_okay = write_data(socket, b"HTTP/1.1 ", status, b"\r\n")
                for header in response_headers:
                    is_okay = is_okay and write_data(socket, header, b"\r\n")
            
                if(body != None): # if body doesn't exist , the handler function is holding the connection and handling it manually 
                    is_okay = is_okay and write_data(socket, b"Content-Length: ", str(len(body)), b"\r\n\r\n")            
                    is_okay = is_okay and write_data(socket, body)
                
                    keep_alive = headers.get("connection", None)
                    if(keep_alive and keep_alive.lower() != "keep-alive"):
                        socket.close()
                        is_okay = False
                else: # close headers and dont rehandle this connection(probably socket if used for sending data in callbacks etc)
                    write_data(socket, b"\r\n")
                    is_okay = False
            else: # return value is None => manually handling
                is_okay = False
        except Exception as ex:
            socket.close()
            server_log(LOG_TYPE_REQUEST_ERROR , cur_millis, data=str(ex))
            if(True):
                import traceback
                traceback.print_exc()
            return
        
        

global_route_handlers = []
def route_handler(regex):
    def _decorator(func):
        global_route_handlers.append((regex, func))

        def new_func(*args, **kwargs):
            return func(*args, **kwargs)

        new_func._original = getattr(func, "_original", func)
        return new_func
    
    return _decorator

    

stream_server = None
def exit_handler():
    server_log(LOG_TYPE_SERVER_INFO,  data="exiting server")
    global is_server_running, stream_server
    
    is_server_running = False
    
    for i in exit_listeners:
        i()
    stream_server.stop()#should handle all connections fgracefully
        #wait for queue to flush/process
        
def start_stream_server(port=80, handlers= [], **ssl_args):
    global_route_handlers.sort(key=functools.cmp_to_key(lambda x, y: len(y[0])-len(x[0]))) #reverse sort by length
    global_route_handlers.extend(handlers)
    for regex, handler in global_route_handlers:
        if(isinstance(regex, str)):
            regex = re.compile(regex)
        request_handlers.append((regex, handler))
        
        
    #request_handlers.sort(key = lambda x:x[0] , reverse=True)
    global stream_server
    stream_server = StreamServer(
    ('', port), handle_connection, **ssl_args)
    
    gevent.signal(signal.SIGINT, exit_handler)

    stream_server.serve_forever()
    
    

##################wsgi related handlers

def application(environ, start_response):
    ret = None
    for handler in request_handlers:
        args = handler[0].match(environ['PATH_INFO'])
        func = handler[1]
        if(args!=None):
            kwargs = {}
            kwargs["query_params"] = urllib.parse.parse_qs(environ['QUERY_STRING'])

            fargs = args.groups()
            if(fargs):
                ret = func(*fargs , **kwargs)
            else:
                ret = func(**kwargs)  
            
            break
        
    status = response_headers = body = None        
    if(ret!=None):
        if(isinstance(ret, tuple)):
            l = len(ret)
            if(l>0):
                body = ret[-1]
            if(l>1):
                response_headers = ret[-2]
            if(l>2):
                status = ret[-3]
        else:
            body = ret
        
        response_headers = response_headers or default_wsgi_headers
        status = status or "200 OK"
        start_response(status, response_headers)
        if(isinstance(body, list)):
            return body
        return [body]
    #close socket
    
    
def start_wsgi_server(port=80, handlers=[]):
    for regex, handler in handlers:
        if(isinstance(regex, str)):
            regex = re.compile(regex)
        request_handlers.append((regex, handler))
        
        
    request_handlers.sort(key = lambda x:x[0] , reverse=True)
    WSGIServer(('', port), application).serve_forever()



'''
Created on 04-Nov-2017

@author: abhinav
'''

import collections
import random
import string
import socket
import struct
from gevent.lock import BoundedSemaphore
import os
from collections import namedtuple
from datetime import datetime
import logging
import time
import hmac
import base64
import hashlib
from gevent.threading import Lock
import re
from _datetime import timezone
import six

from queue import Queue
from threading import Thread
import traceback

from . import config
from .websocket._core import WebSocket
from .config import IS_DEBUG


def cur_ms():
    return int(1000 * time.time())
#random id


max_assumed_sent_buffer_time = 20 * 1000 # milli seconds # TCP_USER_TIMEOUT kernel setting

def set_socket_options(sock):
    l_onoff = 1
    l_linger = 10 # seconds,
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
                 struct.pack('ii', l_onoff, l_linger))# close means a close understand ?
    
    
    
    timeval = struct.pack('ll', 100, 0)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDTIMEO, timeval)

    TCP_USER_TIMEOUT = 18
    sock.setsockopt(socket.SOL_TCP, TCP_USER_TIMEOUT, max_assumed_sent_buffer_time)# close means a close understand ?


class LRUCache:
    def __init__(self, capacity, items=None):
        self.capacity = capacity
        self.cache = collections.OrderedDict(items or [])

    def exists(self, key):
        return self.cache.get(key, None)
        
    def get(self, key, default=None):
        try:
            value = self.cache.pop(key)
            self.cache[key] = value
            return value
        except KeyError:
            return default

    def set(self, key, value):
        removed_entries = []
        try:
            self.cache.pop(key)
        except KeyError:
            while(len(self.cache) >= self.capacity):
                removed_entries.append(self.cache.popitem(last=False))
        self.cache[key] = value
        
        return removed_entries
        
    def delete(self, key):        
        return self.cache.pop(key, None)
    
    
    def to_son(self):
        ret = {}
        for key, val in self.cache:
            ret[key] =   val.to_son() if hasattr(val ,"to_son") else val
        return ret

#get SON of the fields from any generic python object
def to_son(obj):
    ret = obj.__dict__
    for k in ret.keys():
        if(ret[k]==None):
            del ret[k]            
    return ret


def from_kwargs(cls, **kwargs):
    ret = cls()
    for key in kwargs:
        setattr(ret, key, kwargs[key])
    return ret

def get_by_key_list(d , *keyList):
    for key in keyList:
        if(not d):
            return None
        d= d.get(key,None)
    return d

def date2string(date):
    return date.isoformat()

def date2timestamp(dt):
    # datetime to timestamp
    if not isinstance(dt, datetime):
        return dt
    if(six.PY34):
        return dt.replace(tzinfo=timezone.utc).timestamp()
    else:
        timestamp = time.mktime(dt.timetuple()) + dt.microsecond/1e6
        return timestamp

def timestamp2date(timestamp):
    if not isinstance(timestamp, (int, float)):
        return timestamp
    date = datetime.utcfromtimestamp(timestamp)
    return date


###############EPOCH
EPOCH = timestamp2date(0)

def find_index(a_list, value):
    try:
        return a_list.index(value)
    except ValueError:
        return -1



SOME_TIME_WHEN_WE_STARTED_THIS = 1471504855

### this function is inspired from instagram engineering post on generating 64bit keys with 12 bit shard_id inside it
# __thread_data = LRUCache(10000)
def generate_64bit_key(shard_id): # shard id should be a 12 bit number
    # may be later use dattime
    millis_elapsed = int((time.time() - SOME_TIME_WHEN_WE_STARTED_THIS) * 1000)
    
    _id = (((1 << 41) - 1) & millis_elapsed) << 23 # 41 bits , clear 22 places for random id and shard_id
    
    _id |= (((1 << 12) - 1) & shard_id) << 11 # 12 bit shard id, on top
    

    #increment per thread number
    # thread_id = threading.current_thread().ident
    # thread_data = __thread_data.get(thread_id, None)
    # if(not thread_data):
    #     thread_data = [0]
    #     __thread_data.set(thread_id , thread_data)
    # thread_data[0] = (thread_data[0] + 1)%(1 << 11)
    random_11_bits = random.randrange(0, 1 << 11)
    _id |= (((1 << 11) - 1) & random_11_bits) # clear 12 places for random

    #shard_id|timestmap|random_number
    
    return _id

##### protobuf + mysql utility functions
def list_to_protobuf(values, message):
    '''parse list to protobuf message'''
    if not values:
        return
    if isinstance(values[0], dict): # value needs to be further parsed
        for v in values:
            cmd = message.add()
            dict_to_protobuf(v, cmd)
    else: # value can be set
        message.extend(values)


#just like move constructor ;) , move values much faster than copies
def dict_to_protobuf(values, obj, transformations=None, excludes=None, preserve_values=True):
    if(not preserve_values):
        if(transformations):
            for k, func in transformations.items():
                values[k] = func(values.get(k, None))
                
        if(excludes):
            for exclude, flag in excludes.items():
                if(hasattr(values, exclude)):
                    del values[exclude]
             
    for k, v in values.items():
 
        if(preserve_values):
            if(transformations and k in transformations):
                v = transformations[k](v)
                            
            if(excludes and k in excludes):
                continue

        if hasattr(obj, k):
            if isinstance(v, dict): # value needs to be further parsed
                dict_to_protobuf(v, getattr(obj, k))
            elif isinstance(v, list):
                list_to_protobuf(v, getattr(obj, k))
            else: # value can be set
                if v: # otherwise default
                    setattr(obj, k, v)

def get_mysql_rows_as_dict(res):
    
    rows_as_dict = []
    for row in res.rows:
        as_dict = {}
        for field, val in zip(res.fields, row):
            as_dict[field[0]] = val
        
        
        rows_as_dict.append(as_dict)
    
    return rows_as_dict

########################



def get_shard_id(_id):
    return (int(_id) & (((1 << 12) - 1) << 11)) >> 11
    
    
## move it to seperate module ?
## copied from internet sha1 token encode - decode module

if hasattr(hmac, 'compare_digest'):  # python 3.3
    _time_independent_equals = hmac.compare_digest
else:
    def _time_independent_equals(a, b):
        if len(a) != len(b):
            return False
        result = 0
        if isinstance(a[0], int):  # python3 byte strings
            for x, y in zip(a, b):
                result |= x ^ y
        else:  # python2
            for x, y in zip(a, b):
                result |= ord(x) ^ ord(y)
        return result == 0


def utf8(value)->bytes:
    if(isinstance(value, bytes)):
        return value
    return value.encode("utf-8")

def create_signed_value(name, value, secret_key_version="v0"):
    timestamp = utf8(str(int(time.time())))
    value = base64.b64encode(utf8(value))
    secret = config.secrets.get(secret_key_version)
    signature = _create_signature(secret, name, value, timestamp)
    value = b"|".join([value, timestamp, signature, utf8(secret_key_version)])
    return value

def decode_signed_value(name, value, max_age_days=-1):
    if not value:
        return None
    parts = utf8(value).split(b"|")
    secret_key_version = b"v0"
    if len(parts) == 4:
        secret_key_version = parts[3]
    if(len(parts)<3):
        return None
    secret = config.secrets.get(secret_key_version.decode())
    signature = _create_signature(secret, name, parts[0], parts[1])
    if not _time_independent_equals(parts[2], signature):
        return None
    if(max_age_days>0):
        timestamp = int(parts[1])
        if timestamp < time.time() - max_age_days * 86400:
            print (-1, cur_ms(), "Expired cookie %s"%value)
            return None
        if timestamp > time.time() + 31 * 86400:
            # _cookie_signature does not hash a delimiter between the
            # parts of the cookie, so an attacker could transfer trailing
            # digits from the payload to the timestamp without altering the
            # signature.  For backwards compatibility, sanity-check timestamp
            # here instead of modifying _cookie_signature.
            print (-1, cur_ms(), "Cookie timestamp in future; possible tampering %s"%value)
            return None
    if parts[1].startswith(b"0"):
        logging.warning("Tampered cookie %r", value)
        return None
    try:
        return base64.b64decode(parts[0])
    except Exception:
        return None


def _create_signature(secret, *parts):
    hash = hmac.new(utf8(secret), digestmod=hashlib.sha1)
    for part in parts:
        hash.update(utf8(part))
    return utf8(hash.hexdigest())



# Console or Cloud Console.
def get_random_id(length=10, include_timestamp=True):
    '''returns a 10 character random string containing numbers lowercase upper case'''
    '''http://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits-in-python'''

    key_str = ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits+string.ascii_lowercase) for _ in range(length)) 
    if(include_timestamp):
        key_str = key_str + ("%d" % time.time())
    #key_str = hashlib.md5(key_str).hexdigest()
    return key_str

# jump consistent hash function , jump_hash("Asdasd", 100) assigns to a bucket
def jump_hash(key, num_buckets):
    b, j = -1, 0
    key = int(hashlib.md5(key).hexdigest(), 16)
    while j < num_buckets:
        b = int(j)
        key = ((key * int(2862933555777941757)) + 1) & 0xFFFFFFFFFFFFFFFF
        j = float(b + 1) * (float(1 << 31) / float((key >> 33) + 1))
    return int(b)


number_regex = re.compile(r"([0-9\.]+)")
non_alpha_regex = re.compile("[^0-9a-zA-Z \.]", re.DOTALL)
def sanitize_string(text):
    return non_alpha_regex.sub("", text)


non_alpha_regex_2 = re.compile("[^0-9a-zA-Z]", re.DOTALL)
def sanitize_to_id(text):
    return non_alpha_regex_2.sub("", text.lower())


EMAIL_REGEX = re.compile('^[_a-z0-9-]+(\.[_a-z0-9-]+)*(\+[_a-z0-9-]+)?\@[a-z0-9-]+(\.[a-z0-9-]+)*(\.[a-z]{2,4})$')
def is_valid_email(email):
    return EMAIL_REGEX.match(email)
    

def list_diff(first, second):
    if(not first or not second):
        return first
    second = set(second)
    return [item for item in first if item not in second]


class DummyObject:
    entries = None

    def __init__(self, **entries):
        self.entries = entries
        self.__dict__.update(entries)

    def to_dict(self):
        return self.entries

def object_name(o):
  # o.__module__ + "." + o.__class__.__qualname__ is an example in
  # this context of H.L. Mencken's "neat, plausible, and wrong."
  # Python makes no guarantees as to whether the __module__ special
  # attribute is defined, so we take a more circumspect approach.
  # Alas, the module name is explicitly excluded from __qualname__
  # in Python 3.

  module = o.__class__.__module__
  if module is None or module == str.__class__.__module__:
    return o.__class__.__name__  # Avoid reporting __builtin__
  else:
    return module + '.' + o.__class__.__name__

def normalize(val, n=13):
    val = str(val)
    l = len(val)
    if(n >= l):
        return ("0" * (n - l)) + val

    raise Exception('cannot have greater length, ensure about this')

def normalize_no_warn(val, n=13):
    val = str(val)
    l = len(val)
    if(n >= l):
        return ("0" * (n - l)) + val
    return val

def de_normalize(id1):
    i = 0
    l = len(id1)
    while(i < l):
        if(id1[i] != '0'):
            break
        i = i + 1
    if(i >= l ):
        return '0'
    return id1[i:]


class Worker(Thread):
    """ Thread executing tasks from a given tasks queue """
    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            func, args, kargs = self.tasks.get()
            try:
                func(*args, **kargs)
            except Exception as e:
                # An exception happened in this thread
                traceback.print_exc()
            finally:
                # Mark this task as done, whether an exception happened or not
                self.tasks.task_done()

class ThreadPool:
    """ Pool of threads consuming tasks from a queue """
    def __init__(self, num_threads):
        self.tasks = Queue(num_threads)
        for _ in range(num_threads):
            Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        """ Add a task to the queue """
        self.tasks.put((func, args, kargs))

    def map(self, func, args_list):
        """ Add a list of tasks to the queue """
        for args in args_list:
            self.add_task(func, args)

    def wait_completion(self):
        """ Wait for completion of all the tasks in the queue """
        self.tasks.join()

####################################websocket connection and send function , gevent compatible
#gevent _GetPrecompileRelatedFiles
class Connection(WebSocket):
    queue = None# to keep track of how many greenlets are waiting on semaphore to send 
    msg_assumed_sent = None# queue for older sent messages in case of reset we try to retransmit
    ws = None
    lock = BoundedSemaphore()
    is_stale = False
    last_msg_recv_timestamp = None
    last_msg_sent_timestamp = None
    
    
    user_id = None
    is_viewer_list = False
    
    def __init__(self, ws, user_id):
        self.ws = ws
        self.queue = collections.deque()# to keep track of how many greenlets are waiting on semaphore to send 
        self.msg_assumed_sent = collections.deque()# queue for older sent messages in case of reset we try to retransmit
        self.user_id = user_id
   
        self.last_msg_recv_timestamp = self.last_msg_sent_timestamp = time.time()*1000
        
        
    def send(self, msg, ref=None, add_to_assumend_sent=True): # msg is only string data , #ref is used , just in case an exception occurs , we pass that ref 
        if(self.is_stale):
            raise Exception("stale connection")
        
        self.queue.append((ref, msg))
        if(self.lock.locked()):
            return
        
        self.lock.acquire()
        data_ref = None
        data = None
        try:
            while(not self.is_stale and len(self.queue) > 0):
                data_ref, data = self.queue.popleft() #peek
                self.ws.send(data) # msg objects only
                current_timestamp = time.time() * 1000
                self.last_msg_sent_timestamp = current_timestamp
                if(add_to_assumend_sent):
                    while(len(self.msg_assumed_sent) > 0 and self.msg_assumed_sent[0][0] < current_timestamp - max_assumed_sent_buffer_time):
                        #keep inly 100 seconds of previous data
                        self.msg_assumed_sent.popleft()
                    
                    self.msg_assumed_sent.append((current_timestamp , data_ref, data))
                
        except Exception as ex:
            err_msg = "Exception while sending message to %s might be closed "%(self.user_id)
            self.is_stale = True
            raise Exception(err_msg)
            
        finally:
            self.lock.release()
        return
    
    
#dynamo db shit


######upload shit

TypeDescriptor = namedtuple('TypeDescriptor', ['mime_type'])
MIME_TYPE_MAP = dict(
    gif=TypeDescriptor(mime_type='image/gif'),
    webp=TypeDescriptor(mime_type='image/webp'),
    mp4=TypeDescriptor(mime_type='video/mp4'),

    jpeg=TypeDescriptor(mime_type='image/jpeg'),
    jpg=TypeDescriptor(mime_type='image/jpeg'),
    png=TypeDescriptor(mime_type='image/png'),
    json=TypeDescriptor(mime_type='application/json'),
    zip=TypeDescriptor(mime_type='application/zip'),
    mp3=TypeDescriptor(mime_type='audio/mpeg'),
    ogg=TypeDescriptor(mime_type='audio/ogg'),
    opus=TypeDescriptor(mime_type='audio/ogg'),
    wav=TypeDescriptor(mime_type='audio/wav'),
    m4a=TypeDescriptor(mime_type='audio/mp4'),
    aac=TypeDescriptor(mime_type='audio/aac'),
    docx=TypeDescriptor(mime_type='application/vnd.openxmlformats-officedocument.wordprocessingml.document'),
    pdf=TypeDescriptor(mime_type='application/pdf'),

    # variables should not start numerals
    gp=TypeDescriptor(mime_type='audio/3gp'),
    amr=TypeDescriptor(mime_type='audio/amr'),
    wma=TypeDescriptor(mime_type='audio/wma'),
    gpp=TypeDescriptor(mime_type='audio/3gpp'),
    flac=TypeDescriptor(mime_type='audio/flac'),
    
    css=TypeDescriptor(mime_type='text/css'),
    js=TypeDescriptor(mime_type='text/javascript')
    
)

MIME_TYPE_TO_EXTENSION = {v.mime_type: k for k, v in MIME_TYPE_MAP.items()}


def get_mime_type_from_filename(file_name):
    ext = os.path.splitext(file_name)[1][1:]
    return MIME_TYPE_MAP.get(ext, None)













######################gevent handlers

####file handler
def static_file_handler(_base_folder_path_, default_file_path="index.html"):
    cached_file_data = {}
    gevent_lock = Lock()
    def file_handler(sock, path, query_params=None, headers=None, post_data=None):
        if(not path):
            path = default_file_path
        #from given base_path
        path = _base_folder_path_+str(path)
        file_data = cached_file_data.get(path, None)
        resp_headers = None

        if(not file_data or time.time()*1000 - file_data[2] > 1000 if IS_DEBUG else 2*60*1000): #1 millis
            gevent_lock.acquire()
            file_data = cached_file_data.get(path, None)
            if(not file_data or time.time()*1000 - file_data[2] > 1000): #1 millis
                #put a lock here
                try:
                    data = open(path,"rb").read()
                    mime_type_headers = get_mime_type_from_filename(path)
                    if(mime_type_headers):
                        resp_headers = ['Content-Type: %s'%(mime_type_headers.mime_type,)]
                    
                    file_data = (data, resp_headers, time.time()*1000)
                    cached_file_data[path] = file_data
                except Exception as ex:
                    file_data = ("--NO-FILE--", None, time.time()*1000)
                    
            gevent_lock.release()
                    
        data, resp_headers, last_updated = file_data
        return resp_headers, data
    
    return file_handler









from setuptools import setup, find_packages

setup(
	name='blaster-server',
	packages=find_packages("."),
	version='0.0.2',
	license='MIT',
	description='Gevent based python server built from scratch',
	author='Abhinav Reddy',                   # Type in your name
	author_email='abhinavabcd@gmail.com',      # Type in your E-Mail
	url='https://github.com/abhinavabcd/blaster',
	download_url='https://github.com/abhinavabcd/blaster/archive/0.02.tar.gz',
	keywords=['server', 'superfast', 'just like flask but minimal and fast'],
	install_requires=[            # I get to this in a second
					"boto3>=1.4.6",
					"botocore>=1.6.4",
					"pytz>=2017.3",
					"gevent>=1.4.0",
					"greenlet>=0.4.15",
					"pymongo>=3.10.1",
					"elasticsearch>=6.1.1",
					"ujson>=1.35",
					"python-dateutil>=2.6.1",
					"requests>=2.21",
					"requests-aws4auth>=0.9"
	],
	classifiers=[
						'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
						'Intended Audience :: Developers',      # Define that your audience are developers
						'Topic :: Software Development :: Build Tools',
						'License :: OSI Approved :: MIT License',   # Again, pick a license
						'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
						'Programming Language :: Python :: 3.4',
						'Programming Language :: Python :: 3.5',
						'Programming Language :: Python :: 3.6',
	],
)



#! -*- coding: utf-8 -*-


import pprint

from decimal import Decimal
from botocore.exceptions import ClientError
from botocore.vendored.requests.exceptions import ConnectionError

from .helpers import get_attribute_type
from .errors import ClientException, ConnectionException, ParameterException

pp = pprint.PrettyPrinter(indent=4)
pprint = pp.pprint

__all__ = ['Table']


class Table(object):

    def __init__(self, instance, db):
        self.instance = instance
        self.db=db
        self.table_name = instance.__table_name__
        self.table = db.Table(self.table_name)

    def info(self):
        try:
            response = self.db.meta.client.describe_table(TableName=self.table_name)
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])
        else:
            table_info = response['Table']
        return table_info

    def _prepare_hash_key(self):
        hash_key = self.instance._hash_key
        param = {
            'AttributeName': hash_key,
            'KeyType': 'HASH'
        }
        return param

    def _prepare_range_key(self, range_key=None):
        if not range_key:
            range_key = self.instance._range_key
        if range_key:
            param = {
                'AttributeName': range_key,
                'KeyType': 'RANGE'
            }
            return param
        return {}

    def _prepare_key_schema(self):
        KeySchema = []
        hash_key_param = self._prepare_hash_key()
        KeySchema.append(hash_key_param)
        range_key_param = self._prepare_range_key()
        if range_key_param:
            KeySchema.append(range_key_param)
        return KeySchema

    def _prepare_attribute_definitions(self):
        AttributeDefinitions = []
        attributes = self.instance.attributes
        hash_key = self.instance._hash_key
        _fields_indexed = {}
        AttributeDefinitions.append({
            'AttributeName': hash_key,
            'AttributeType': get_attribute_type(attributes[hash_key]),
        })
        _fields_indexed[hash_key] = True
        range_key = self.instance._range_key
        if range_key:
            AttributeDefinitions.append({
                'AttributeName': range_key,
                'AttributeType': get_attribute_type(attributes[range_key]),
            })
            _fields_indexed[range_key] = True
            
        for field in self.instance._local_indexed_fields:
            AttributeDefinitions.append({
                'AttributeName': field,
                'AttributeType': get_attribute_type(attributes[field]),
            })
            _fields_indexed[field] = True


        #for global indexes
        for index_name, attribute_dict in self.instance._global_indexes.items():
            field = attribute_dict["hash_key"]
            if(not _fields_indexed.get(field, None)):
                AttributeDefinitions.append({
                    'AttributeName': field,
                    'AttributeType': get_attribute_type(attributes[field])
                })
            range_key = attribute_dict.get("range_key",None)
            if(range_key and not _fields_indexed.get(range_key, None)):
                AttributeDefinitions.append({
                    'AttributeName': range_key,
                    'AttributeType': get_attribute_type(attributes[range_key])
                })


                    
        return AttributeDefinitions

    def _prepare_primary_key(self, params):
        params['KeySchema'] = self._prepare_key_schema()
        params['AttributeDefinitions'] = self._prepare_attribute_definitions()
        return params

    def _prepare_local_indexes(self):
        indexes = []
        for field in self.instance._local_indexed_fields:
            index_name = '{table_name}_ix_{field}'.format(
                table_name=self.table_name, field=field)
            KeySchema = [self._prepare_hash_key()]
            range_key_param = self._prepare_range_key(field)
            if range_key_param:
                KeySchema.append(range_key_param)
            
            index_properties = {
                'IndexName': index_name,
                'KeySchema': KeySchema,
            }
            field_object = self.instance.attributes[field]
            if(field_object.projections):
                index_properties['Projection'] = {'ProjectionType': 'INCLUDE',
                                                  'NonKeyAttributes': field_object.projections                                                  
                                                 }
            else:
                index_properties['Projection'] = {'ProjectionType': 'KEYS_ONLY'}

            indexes.append(index_properties)
        return indexes

    def _prepare_global_indexes(self):
        indexes = []
        for index_name, attribute_dict in self.instance._global_indexes.items():
            KeySchema = [{
                'AttributeName': attribute_dict["hash_key"], 
                'KeyType': 'HASH'
            }]
            if attribute_dict.get("range_key",None):
                KeySchema.append({
                        'AttributeName': attribute_dict.get("range_key"), 
                        'KeyType': 'RANGE'
                })
            index_properties = {
                'IndexName': index_name,
                'KeySchema': KeySchema,
                'ProvisionedThroughput': {
                    'ReadCapacityUnits': attribute_dict.get("read_capacity"),
                    'WriteCapacityUnits': attribute_dict.get("write_capacity")
                }
            }
            
            if("projections" in attribute_dict):
                index_properties['Projection'] = {'ProjectionType': 'INCLUDE',
                                                  'NonKeyAttributes': attribute_dict["projections"]
                                                 }
            else:
                index_properties['Projection'] = {'ProjectionType': 'KEYS_ONLY'}

            
            indexes.append(index_properties)
        return indexes

    def _prepare_create_table_params(self):
        # TableName
        table_params = {
            'TableName': self.table_name
        }
        # KeySchema && AttributeDefinitions
        table_params = self._prepare_primary_key(table_params)
        # LocalSecondaryIndexes
        local_indexes = self._prepare_local_indexes()
        if local_indexes:
            table_params['LocalSecondaryIndexes'] = local_indexes
        # GlobalSecondaryIndexes
        global_indexes = self._prepare_global_indexes()
        if global_indexes:
            table_params['GlobalSecondaryIndexes'] = global_indexes
        # ProvisionedThroughput
        table_params['ProvisionedThroughput'] = {
            'ReadCapacityUnits': self.instance.ReadCapacityUnits,
            'WriteCapacityUnits': self.instance.WriteCapacityUnits
        }
        return table_params

    def create(self):
        '''
        # create table
        create_table Request Syntax
        # http://boto3.readthedocs.io/en/sinstance/reference/services/dynamodb.html#DynamoDB.Client.create_instance
        response = client.create_table(
            AttributeDefinitions=[
                {
                    'AttributeName': 'string',
                    'AttributeType': 'S'|'N'|'B'
                },
            ],
            TableName='string',
            KeySchema=[
                {
                    'AttributeName': 'string',
                    'KeyType': 'HASH'|'RANGE'
                },
            ],
            LocalSecondaryIndexes=[
                {
                    'IndexName': 'string',
                    'KeySchema': [
                        {
                            'AttributeName': 'string',
                            'KeyType': 'HASH'|'RANGE'
                        },
                    ],
                    'Projection': {
                        'ProjectionType': 'ALL'|'KEYS_ONLY'|'INCLUDE',
                        'NonKeyAttributes': [
                            'string',
                        ]
                    }
                },
            ],
            GlobalSecondaryIndexes=[
                {
                    'IndexName': 'string',
                    'KeySchema': [
                        {
                            'AttributeName': 'string',
                            'KeyType': 'HASH'|'RANGE'
                        },
                    ],
                    'Projection': {
                        'ProjectionType': 'ALL'|'KEYS_ONLY'|'INCLUDE',
                        'NonKeyAttributes': [
                            'string',
                        ]
                    },
                    'ProvisionedThroughput': {
                        'ReadCapacityUnits': 123,
                        'WriteCapacityUnits': 123
                    }
                },
            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 123,
                'WriteCapacityUnits': 123
            },
            StreamSpecification={
                'StreamEnabled': True|False,
                'StreamViewType': 'NEW_IMAGE'|'OLD_IMAGE'|'NEW_AND_OLD_IMAGES'|'KEYS_ONLY'
            }
        )

        AttributeType (string) -- [REQUIRED]
            The data type for the attribute, where:

            * S - the attribute is of type String
            * N - the attribute is of type Number
            * B - the attribute is of type Binary
        KeySchema (list) -- [REQUIRED]
        KeyType - The role that the key attribute will assume:
            * HASH - partition key
            * RANGE - sort key
        '''
        try:
            params = self._prepare_create_table_params()
            return self.db.create_table(**params)
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])
        except ConnectionError:
            raise ConnectionException('Connection refused')

    def _update_throughput(self, ProvisionedThroughput):
        ReadCapacityUnits = ProvisionedThroughput['ReadCapacityUnits']
        WriteCapacityUnits = ProvisionedThroughput['WriteCapacityUnits']
        if (ReadCapacityUnits != self.instance.ReadCapacityUnits or
                WriteCapacityUnits != self.instance.WriteCapacityUnits):
            self.table.update(ProvisionedThroughput={
                'ReadCapacityUnits': self.instance.ReadCapacityUnits,
                'WriteCapacityUnits': self.instance.WriteCapacityUnits
            })

    def _update_streams(self):
        # TODO
        pass

    def _update_global_indexes(self):
        # TODO
        pass

    def update(self):
        '''
        # update table
        http://boto3.readthedocs.io/en/stable/reference/services/dynamodb.html#DynamoDB.Table.update
        You can only perform one of the following operations at once:

        * Modify the provisioned throughput settings of the table.
        * Enable or disable Streams on the table.
        * Remove a global secondary index from the table.
        * Create a new global secondary index on the table.
          Once the index begins backfilling, you can use UpdateTable to perform
          other operations.

        UpdateTable is an asynchronous operation; while it is executing,
        the table status changes from ACTIVE to UPDATING. While it is UPDATING,
        you cannot issue another UpdateTable request.
        When the table returns to the ACTIVE state, the UpdateTable operation is
        complete.

        # Request Syntax

        {
           "AttributeDefinitions": [
              {
                 "AttributeName": "string",
                 "AttributeType": "string"
              }
           ],
           "GlobalSecondaryIndexUpdates": [
              {
                 "Create": {
                    "IndexName": "string",
                    "KeySchema": [
                       {
                          "AttributeName": "string",
                          "KeyType": "string"
                       }
                    ],
                    "Projection": {
                       "NonKeyAttributes": [ "string" ],
                       "ProjectionType": "string"
                    },
                    "ProvisionedThroughput": {
                       "ReadCapacityUnits": number,
                       "WriteCapacityUnits": number
                    }
                 },
                 "Delete": {
                    "IndexName": "string"
                 },
                 "Update": {
                    "IndexName": "string",
                    "ProvisionedThroughput": {
                       "ReadCapacityUnits": number,
                       "WriteCapacityUnits": number
                    }
                 }
              }
           ],
           "ProvisionedThroughput": {
              "ReadCapacityUnits": number,
              "WriteCapacityUnits": number
           },
           "StreamSpecification": {
              "StreamEnabled": boolean,
              "StreamViewType": "string"
           },
           "TableName": "string"
        }
        '''
        table_info = self.info()
        ProvisionedThroughput = table_info['ProvisionedThroughput']
        self._update_throughput(ProvisionedThroughput)

    def delete(self):
        # delete table
        try:
            return self.table.delete()
        except ClientError:
            raise ClientException('Cannot do operations on a non-existent table')
        except ConnectionError:
            raise ConnectionException('Connection refused')

    def _get_primary_key(self, **kwargs):
        hash_key, range_key = self.instance._hash_key, self.instance._range_key
        hash_value = kwargs.get(hash_key) or getattr(self.instance, hash_key)
        if isinstance(hash_value, (int, float)):
            hash_value = Decimal(hash_value)
        key = {
            hash_key: hash_value
        }
        if(range_key):
            range_value = kwargs.get(range_key) or getattr(self.instance, range_key, None)
            if not range_value:
                raise ParameterException('Invalid range key value type')
            if isinstance(range_value, (int, float)):
                range_value = Decimal(range_value)
            key[range_key] = range_value
        return key

    def get_item(self, **kwargs):
        """
        primary_key: params: primary_key dict
        """
        kwargs['Key'] = kwargs.get('Key') or self._get_primary_key()
        try:
            response = self.table.get_item(**kwargs)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ValidationException':
                return None
            raise ClientException(e.response['Error']['Message'])
        else:
            item = response.get('Item')
        return item

    def batch_get_item(self, *primary_keys, **kwargs):
        """
        primary_key: params: primary_keys list
        """
        _primary_keys = []
        for primary_key in primary_keys:
            key = self._get_primary_key(**primary_key)
            _primary_keys.append(key)
        params = {
            'RequestItems': {
                self.table_name: {
                    'Keys': _primary_keys
                }
            },
            'ReturnConsumedCapacity': 'TOTAL'
        }
        projections = kwargs.get("projections",None)
        if(projections):
            params["RequestItems"]["ProjectionExpression"] = ",".join(projections)
        
        try:
            response = self.db.batch_get_item(**params)
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])
        else:
            items = response['Responses'][self.table_name]
        return items

    def put_item(self, item, overwrite=True):
        args = dict(Item=item)
        if(not overwrite):
            expr = 'attribute_not_exists(%s)'%(self.instance.__class__._hash_key,)
            if(self.instance.__class__._range_key):
                expr += (" AND attribute_not_exists(%s)"%(self.instance.__class__._range_key,))
            args["ConditionExpression"] = expr
            
            
        self.table.put_item(**args)
        return True

    def batch_write(self, items, overwrite=False):
        pkeys = []
        if overwrite:
            instance = self.instance
            pkeys = [instance._hash_key, instance._range_key]
        try:
            with self.table.batch_writer(overwrite_by_pkeys=pkeys) as batch:
                for item in items:
                    batch.put_item(Item=item)
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])
        
    def batch_delete(self, items=None, keys=None):
        try:
            with self.table.batch_writer() as batch:
                if(items):
                    for item in items:
                        batch.delete_item(Key=Table(item)._get_primary_key())
                if(keys):
                    for key in keys:
                        batch.delete_item(Key=key)
                    
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])


    def query(self, **kwargs):
        """
        response = table.query(
            IndexName='string',
            Select='ALL_ATTRIBUTES'|'ALL_PROJECTED_ATTRIBUTES'|'SPECIFIC_ATTRIBUTES'|'COUNT',
            Limit=123,
            ConsistentRead=True|False,
            ScanIndexForward=True|False,
            ExclusiveStartKey={
                'string': 'string'|123|Binary(b'bytes')|True|None|set(['string'])|set([123])|set([Binary(b'bytes')])|[]|{}
            },
            ReturnConsumedCapacity='INDEXES'|'TOTAL'|'NONE',
            ProjectionExpression='string',
            FilterExpression=Attr('myattribute').eq('myvalue'),
            KeyConditionExpression=Key('mykey').eq('myvalue'),
            ExpressionAttributeNames={
                'string': 'string'
            },
            ExpressionAttributeValues={
                'string': 'string'|123|Binary(b'bytes')|True|None|set(['string'])|set([123])|set([Binary(b'bytes')])|[]|{}
            }
        )
        ExclusiveStartKey: 起始查询的key，也就是上一页的最后一条数据
        ConsistentRead: 是否使用强制一致性 默认False
        ScanIndexForward: 索引的排序方式 True 为正序 False 为倒序 默认True
        ReturnConsumedCapacity: DynamoDB 写入期间使用的写入容量单位
            TOTAL 会返回由表及其所有global secondary index占用的写入容量；
            INDEXES 仅返回由global secondary index占用的写入容量；
            NONE 表示您不需要返回任何占用容量统计数据。
        ProjectionExpression: 用于指定要在扫描结果中包含的属性
        FilterExpression: 指定一个条件，以便仅返回符合条件的项目
        KeyConditionExpression: 要查询的键值
        ExpressionAttributeNames: 提供名称替换功能
        ExpressionAttributeValues: 提供值替换功能
        """
        try:
            #print(kwargs)
            response = self.table.query(**kwargs)
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])
        return response

    def scan(self, **kwargs):
        try:
            #print(kwargs)
            response = self.table.scan(**kwargs)
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])
        return response

    def _prepare_update_item_params(self, update_fields=None, *args, **kwargs):
        params = {
            'Key': self._get_primary_key()
        }
        _condition_expression = kwargs.pop('ConditionExpression', None)
        if _condition_expression:
            params['ConditionExpression'] = _condition_expression
        _expression_attribute_values = kwargs.pop('ExpressionAttributeValues', {})
        _expression_attribute_names =  kwargs.pop('ExpressionAttributeNames', {})
        action_exp_dict = {}
        if update_fields:
            set_expression_str = ''
            for k, v in update_fields.items():
                label = ':{k}'.format(k=k)
                path = '#{k}'.format(k=k)
                if set_expression_str:
                    set_expression_str += ', {k} = {v}'.format(k=path, v=label)
                else:
                    set_expression_str += '{k} = {v}'.format(k=path, v=label)
                _expression_attribute_values[label] = v
                _expression_attribute_names[path] = k
            action_exp_dict['SET'] = set_expression_str
        for arg in args:
            exp, exp_attr, action = arg
            eav = exp_attr.get('value', {})
            ean = exp_attr.get('name', {})
            action_exp = action_exp_dict.get(action)
            if action_exp:
                action_exp = '{action_exp}, {exp}'.format(action_exp=action_exp,
                                                          exp=exp)
            else:
                action_exp = exp
            action_exp_dict[action] = action_exp
            _expression_attribute_values.update(eav)
            _expression_attribute_names.update(ean)
        for action, _exp in action_exp_dict.items():
            action_exp_dict[action] = '{action} {exp}'.format(action=action,
                                                              exp=_exp)
        if _expression_attribute_values:
            params['ExpressionAttributeValues'] = _expression_attribute_values
        if _expression_attribute_names:
            params['ExpressionAttributeNames'] = _expression_attribute_names
        params['UpdateExpression'] = " ".join(action_exp_dict.values())
        params.update(kwargs)
        return params

    def update_item(self, update_fields, *args, **kwargs):
        '''
        update_fields: update_fields (dict)
        http://boto3.readthedocs.io/en/stable/reference/services/dynamodb.html#DynamoDB.Table.update_item
        response = table.update_item(
            Key={
                'string': 'string'|123|Binary(b'bytes')|True|None|set(['string'])|set([123])|set([Binary(b'bytes')])|[]|{}
            },
            ReturnValues='NONE'|'ALL_OLD'|'UPDATED_OLD'|'ALL_NEW'|'UPDATED_NEW',
            ReturnConsumedCapacity='INDEXES'|'TOTAL'|'NONE',
            ReturnItemCollectionMetrics='SIZE'|'NONE',
            UpdateExpression='string',
            ConditionExpression=Attr('myattribute').eq('myvalue'),
            ExpressionAttributeNames={
                'string': 'string'
            },
            ExpressionAttributeValues={
                'string': 'string'|123|Binary(b'bytes')|True|None|set(['string'])|set([123])|set([Binary(b'bytes')])|[]|{}
            }
        )
        ## example
        item.update_item(a=12, b=12, c=12)
        '''
        params = self._prepare_update_item_params(update_fields, *args, **kwargs)
        try:
            item = self.table.update_item(**params)
            attributes = item.get('Attributes')
            return attributes
        except ClientError as e:
            if e.response['Error']['Code'] == "ConditionalCheckFailedException":
                print(e.response['Error']['Message'])
            raise ClientException(e.response['Error']['Message'])

    def _prepare_delete_item_params(self, **kwargs):
        params = {
            'Key': self._get_primary_key()
        }
        _condition_expression = kwargs.pop('ConditionExpression', None)
        if _condition_expression:
            params['ConditionExpression'] = _condition_expression
        _expression_attribute_values = kwargs.pop('ExpressionAttributeValues', {})
        if _expression_attribute_values:
            params['ExpressionAttributeValues'] = _expression_attribute_values
        _expression_attribute_names =  kwargs.pop('ExpressionAttributeNames', {})
        if _expression_attribute_names:
            params['ExpressionAttributeNames'] = _expression_attribute_names
        return params

    def delete_item(self, key=None, **kwargs):
        '''
        http://boto3.readthedocs.io/en/stable/reference/services/dynamodb.html#DynamoDB.Table.delete_item
        Deletes a single item in a table by primary key. You can perform a
        conditional delete operation that deletes the item if it exists,
        or if it has an expected attribute value.

        In addition to deleting an item, you can also return the item's
        attribute values in the same operation, using the ReturnValues parameter.

        Unless you specify conditions, the DeleteItem is an idempotent operation;
        running it multiple times on the same item or attribute does not result
        in an error response.

        Conditional deletes are useful for deleting items only if specific
        conditions are met. If those conditions are met, DynamoDB performs the
        delete. Otherwise, the item is not deleted.

        Request Syntax

        response = table.delete_item(
            Key={
                'string': 'string'|123|Binary(b'bytes')|True|None|set(['string'])|set([123])|set([Binary(b'bytes')])|[]|{}
            },
            ReturnValues='NONE'|'ALL_OLD'|'UPDATED_OLD'|'ALL_NEW'|'UPDATED_NEW',
            ReturnConsumedCapacity='INDEXES'|'TOTAL'|'NONE',
            ReturnItemCollectionMetrics='SIZE'|'NONE',
            ConditionExpression=Attr('myattribute').eq('myvalue'),
            ExpressionAttributeNames={
                'string': 'string'
            },
            ExpressionAttributeValues={
                'string': 'string'|123|Binary(b'bytes')|True|None|set(['string'])|set([123])|set([Binary(b'bytes')])|[]|{}
            }
        )
        Parameters:
            Key (dict) -- [REQUIRED]
        '''
        key = key or self._get_primary_key()
        try:
            self.table.delete_item(Key=key, ReturnValues='ALL_OLD')
        except ClientError as e:
            if e.response['Error']['Code'] == "ConditionalCheckFailedException":
                raise ClientException(e.response['Error']['Message'])
        return True

    def item_count(self):
        return self.table.item_count



'''
Created on 04-Nov-2017

@author: abhinav
'''
from os import environ



secrets = {"v0": "something_random_secret_here"}

IS_DEBUG = True
if(environ.get('PROD') == "1"):
    IS_DEBUG = False



aws_config = {
    'aws_access_key_id': "A*****",
    'aws_secret_access_key': "G*****",
    'region_name': "ap-south-1"
}

es_http_host = None
es_aws_host = None

connection_generators = {}

sqs_url = None



'''
Created on May 11, 2016

@author: abhinav
'''
import urllib.request
import urllib.error
import urllib.parse
import zlib
from io import StringIO, BytesIO
import hashlib
import os
from queue import Queue
import time
import ujson as json
import six


url_loaders_queue = Queue()

class OpenerDirector2(urllib.request.OpenerDirector):
    def __init__(self):
        super().__init__()
        self.addheaders = [] # remove default python-urllib-header


urllib.request.OpenerDirector = OpenerDirector2 # override for everyone

def get_url_loader(actual_func):
    
    def ret_func(*args, **kwargs):
        url_loader = None
        try:
            url_loader = url_loaders_queue.get(block=False)
        except Exception as ex:
            http_logger = urllib.request.HTTPHandler(debuglevel=0)
            https_logger = urllib.request.HTTPSHandler(debuglevel=0)
            url_loader = urllib.request.build_opener(http_logger, urllib.request.HTTPCookieProcessor(), urllib.request.ProxyHandler(), https_logger, urllib.request.HTTPRedirectHandler())
            urllib.request.install_opener(url_loader)
        
        if(kwargs == None):
            kwargs = {}
        kwargs["url_loader"] = url_loader
        ret = actual_func(*args, **kwargs)
        url_loaders_queue.put(url_loader)
        
        return ret
        
    return ret_func


'''returns Io'''
@get_url_loader
def get_data(url, post=None, headers={}, method=None, url_loader=None, as_string_buffer=True):
    headers['Accept-encoding'] = 'gzip'
    ret = None
    try:
        if(post and isinstance(post, six.text_type)):
            post = post.encode()
        req = urllib.request.Request(url, post, headers)
        if(method != None):
            req.get_method = lambda : method
        ret = url_loader.open(req)
        if ret.info().get('Content-Encoding') == 'gzip':
            ret2 = ret
            try:
                ret = BytesIO(zlib.decompress(ret2.read(), 16 + zlib.MAX_WBITS))
            except Exception as ex:
                decompressor = zlib.decompressobj()
                ret = BytesIO(decompressor.decompress(ret2.read()))
            ret2.close()
            
    except urllib.error.HTTPError as e:
        ret = None
        err_body = e.read()
        print(-4 , int(time.time() * 1000), json.dumps({"err": str(e), "body": err_body.decode() if err_body else ""}))
    except Exception as e:
        ret = None
        print(-5 , int(time.time() * 1000), json.dumps({"err": str(e)}))
    if(as_string_buffer):
        return StringIO(ret.read().decode())
    return ret


def get_data_cached(url, post=None, headers={}, method=None, cache=True, ignore_cache_read=False, cache_folder="/tmp/", as_string_buffer=True):
    cache_hash = url
    if(post):
        cache_hash += json.dumps(post)
    cache_hash = hashlib.md5(cache_hash.encode("utf-8")).hexdigest()
    if(cache and not ignore_cache_read):
        if(os.path.isfile(cache_folder + cache_hash)):
            #print "reading from cache"
            bytes_buffer = BytesIO(open(cache_folder + cache_hash, "rb").read())
            if(as_string_buffer):
                return StringIO(bytes_buffer.read().decode())
            else:
                return bytes_buffer

    _resp = get_data(url, post=post, headers=headers, as_string_buffer=False)
    if(_resp):
        _data = _resp.read()
        if(cache):
            open(cache_folder + cache_hash, "wb").write(_data)
        if(as_string_buffer):
            return StringIO(_data.decode())
        return BytesIO(_data)
    return None
        
            

if __name__ == "__main__":
    for i in range(10):
        print((get_data("http://WWW.google.com")))



'''
Created on 22-Aug-2017

@author: abhinav
'''
import random
from blaster.base import start_stream_server, route_handler,\
    process_post_params, process_post_data
from blaster.common_funcs_and_datastructures import static_file_handler
import os


@route_handler("^/test$") 
def do_test(sock ,query_params=None, headers=None, post_data=None):
    return "<html><body>Hello World</body></html>"


##directly return an object , server will automatically convert to json
@route_handler("^/say/(.*)$") 
def do_test2(sock, something, query_params=None, headers=None, post_data=None):
    return {"message" : str(something)} #something is a byte array !sorry

##directly return an object , server will automatically convert to json
@route_handler("^/echo_params/(.*)$")
def do_test3(sock, something, query_params=None, headers=None, post_data=None):
    return {"query_params" : query_params}

##FOR handling you post arguments as key value data
@route_handler("^/echo_post_params")
@process_post_params
def do_test4(sock, query_params=None, headers=None, post_data=None):
    return {"query_params" : query_params}

##FOR handling you post as raw body
@route_handler("^/echo_post_data")
def do_test5(sock, query_params=None, headers=None, post_data=None):
    return {"post_data" : post_data}

##FOR handling you post as raw json body
@route_handler("^/echo_post_data")
@process_post_data
def do_test6(sock, query_params=None, headers=None, post_data=None):
    return {"post_data_object" : query_params}

#respond custom headers
@route_handler("^/custom_header$")
def do_test7(sock, query_params=None, headers=None, post_data=None):
    return ["X-header: something_random"], {"post_data_object" : query_params}






if __name__ == "__main__":    
    start_stream_server(8081, handlers=[
            ('^/(.*)', static_file_handler(os.path.dirname(os.path.abspath(__file__))+"/web/"))  #can server files too                  
    ])


#! -*- coding: utf-8 -*-
'''
All the errors specific to Dynamodb.  The goal is to mimic ActiveRecord.
'''


__all__ = ['ClientException', 'ConnectionException', 'ParameterException',
           'FieldValidationException', 'ValidationException']


class ValidationException(Exception):
    pass


class ClientException(Exception):
    pass


class ConnectionException(Exception):
    pass


class ParameterException(Exception):
    pass


class FieldValidationException(Exception):

    def __init__(self, errors, *args, **kwargs):
        super(FieldValidationException, self).__init__(*args, **kwargs)
        self._errors = errors

    @property
    def errors(self):
        return self._errors

    def __str__(self):
        return self._errors


class NotFoundError(Exception):

    def __init__(self, errors, *args, **kwargs):
        super(NotFoundError, self).__init__(*args, **kwargs)
        self._errors = errors

    @property
    def errors(self):
        return self._errors

    def __str__(self):
        return self._errors
